/*++

Copyright (c) Sydus.

Module:

    SSConnection.h

Authors:

    Sharvil Nanavati (sharvil) 2008-04-17

--*/

#pragma once

#include "base.h"
#include "NetLib.h"

#include <list>
#include <map>

class ICryptosystem;

class SSConnection
{
	public:
		friend class SSConnectionPool;
		enum State
		{
			KIdle,
			KKeyExchange,
			KConnected,
			KDisconnected
		};

		void Disconnect();

		State GetState() const;
		std::string GetName() const;
		InetAddress GetRemoteAddress() const;
		uint32 ToRemoteTimestamp(uint32 localTimestamp) const;

		bool IsContinuum() const;

		void Send(const BString & packet, bool reliable = false);
		void SendFile(const BString & contents);
		void SendFile(const std::string & filename);

		void SynchronizeClocks();
		void RequestExecutableChecksum(uint32 key);

		bool HasMorePackets() const;
		BString GetNextPacket();

	protected:
		//
		// The lifetime of this object is governed by the connection pool.
		//
		explicit SSConnection(const std::string & name, const InetAddress & remoteAddr, UdpServer & socket, const std::string & oracleHost = "sharvil.nanavati.net", uint16 oraclePort = 6000);
		~SSConnection();

		void ConnectSubSpace();
		void ConnectContinuum();

		//
		// These are used by the connection pool to read/write our packet
		// queues.
		//
		void AppendInboundPacket(const BString & packet);
		bool HasMoreOutboundPackets() const;
		BString GetNextOutboundPacket();

		//
		// Used by the connection pool to do resends and update state.
		//
		void Poll();

	private:
		static const uint32 KConnectTimeout;
		static const uint32 KMaximumEcpVersion;
		static const uint32 KDefaultWindowSize;

		typedef DataReader <LittleEndian> reader_t;
		typedef DataWriter <LittleEndian> writer_t;

		uint32 GetResendTime() const;

		void RawSend(const BString & packet, bool resend);
		void SendImmediately(const BString & packet);
		void ProcessPacket(const BString & packet);
		void EncryptPacket(BString & packet);
		void DecryptPacket(BString & packet);

		void RequestConnection(uint8 encryptionMode);

		void ProcessConnectionRequest(const BString & packet);
		void ProcessSubSpaceKey(const BString & packet);
		void ProcessSequencePacket(const BString &  packet);
		void ProcessSequenceAck(const BString & packet);
		void ProcessClockSyncRequest(const BString & packet);
		void ProcessClockSyncResponse(const BString & packet);
		void ProcessLargePacket(const BString & packet);
		void ProcessHugePacket(const BString & packet);
		void ProcessHugePacketCancelAck(const BString & packet);
		void ProcessClusteredPacket(const BString & packet);
		void ProcessContinuumKey(const BString & packet);
		void ProcessOracleKeyReply(const BString & packet);
		void ProcessOracleChecksumReply(const BString & packet);
		void ProcessUpgradeProtocolRequest(const BString & packet);
		void ProcessUpgradeProtocolReply(const BString & packet);
		void ProcessAcknowlegeAll(const BString & packet);
		void ProcessSetWindowSize(const BString & packet);

		bool IsSequencePacket(const BString & packet) const;
		bool IsAckPacket(const BString & packet) const;

		void SendSetWindowSize();

		typedef std::list <BString> PacketQueue;
		typedef std::list <std::pair <BString, bool> > DeferredCall;
		typedef std::map <uint32, BString> SequenceQueue;
		typedef std::multimap <uint32, BString> ResendQueue;

		State iState;
		std::string iName;
		InetAddress iRemoteAddress;

		UdpServer & iSocket;
		UdpClient iOracleSocket;
		ICryptosystem * iCrypt;

		PacketQueue iInboundPackets;
		PacketQueue iOutboundPackets;

		DeferredCall iDeferredSends;

		SequenceQueue iSequencePackets;
		uint32 iRemoteSequence;
		ResendQueue iResendPackets;
		uint32 iLocalSequence;

		BString iLargePacket;
		BString iHugePacket;

		bool iIsContinuum;

		uint32 iClockDifference;
		uint32 iPacketsSent;
		uint32 iPacketsReceived;

		uint32 iConnectTime;
		uint32 iEcpVersion;
		uint32 iRemoteWindowSize;
};
