/*++

Copyright (c) Sharvil Nanavati.

Module:

    main.cpp

Authors:

    Sharvil Nanavati (sharvil) 2008-04-27

--*/

#include "base.h"
#include "SSNetLib.h"
#include "TimeLib.h"
#include "LogLib.h"
#include "Config.h"
#include "ZoneList.h"

#include <algorithm>

#if !defined(PLATFORM_WIN32)
#	include <signal.h>
#endif

bool gQuitSignal = false;

#if defined(PLATFORM_WIN32)
	BOOL WINAPI quitHandler(DWORD type)
	{
		gQuitSignal = true;
		return TRUE;
	}

	void setSignalHandler()
	{
		SetConsoleTitle("SubSpace Directory Server");
		SetConsoleCtrlHandler(quitHandler, true);
	}
#else
	void quitHandler(int32)
	{
		gQuitSignal = true;
	}

	void setSignalHandler()
	{
		struct sigaction quitSignal;
		memset(&quitSignal, 0, sizeof(quitSignal));
		quitSignal.sa_handler = quitHandler;
		if(sigaction(SIGINT, &quitSignal, 0) < 0)
			throw sydus_error("Unable to set signal handler.");
	}
#endif

int main(int argc, const_cstring * argv)
{
	try
	{
		setSignalHandler();

		Config config;
		Time launchTime;

		SSConnectionPool pool;
		UdpServer updateServer;
		pool.Bind(config.GetBaseListenPort());
		updateServer.Bind(config.GetBaseListenPort() + 1);

		Selector selector;
		updateServer.Register(selector, Selector::OP_READ);

		ZoneList zones(config);
		StringISet upstreamServers = config.GetUpstreamServerList();
		for(StringISet::const_iterator i = upstreamServers.begin(); i != upstreamServers.end(); ++i)
			zones.AddSource(InetAddress(*i));

		Logger::Instance().Log(KLogInfo, " -- Snrrrub's Directory Server 1.0 started --");
		printf(" -- Snrrrub's Directory Server 1.0 started --\n");
		printf("Press CTRL+C to exit.\n");

		Stopwatch configTimer;
		Stopwatch statsTimer;

		configTimer.Start();
		statsTimer.Start();
		while(!gQuitSignal)
		{
			selector.Select(100);

			if(configTimer.GetTimeElapsed() / 1000 >= config.GetConfigReloadPeriod())
			{
				if(config.Load())
				{
					typedef std::vector <std::string> sv;
					Logger::Instance().Log(KLogInfo, "* Configuration reloaded.");

					StringISet newUpstream = config.GetUpstreamServerList();

					sv difference(std::max(upstreamServers.size(), newUpstream.size()));
					sv::iterator end = std::set_difference(upstreamServers.begin(), upstreamServers.end(), newUpstream.begin(), newUpstream.end(), difference.begin());
					for(sv::iterator i = difference.begin(); i != end; ++i)
						zones.RemoveSource(InetAddress(*i));

					end = std::set_difference(newUpstream.begin(), newUpstream.end(), upstreamServers.begin(), upstreamServers.end(), difference.begin());
					for(sv::iterator i = difference.begin(); i != end; ++i)
						zones.AddSource(InetAddress(*i));

					upstreamServers = newUpstream;
				}
				configTimer.Reset();
			}

			if(!selector.ReadSet().empty())
			{
				InetAddress source;
				BString packet = updateServer.Read(source, 1024);
				if(packet.GetLength() == 4)
				{
					BString pong("\x01\x00\x00\x00", 4);
					pong.Append(packet);
					updateServer.Write(source, pong);
				}
				else
					zones.ServerUpdate(source, packet);
			}

			pool.Poll();
			for(SSConnectionPool::iterator i = pool.begin(); i != pool.end(); ++i)
				while(i->second->HasMorePackets())
					zones.ClientRequest(*i->second, i->second->GetNextPacket());

			zones.Poll();

			if(statsTimer.GetTimeElapsed() / 1000 >= config.GetStatisticsDumpPeriod())
			{
				zones.WriteStatistics(launchTime);
				statsTimer.Reset();
			}
		}

		printf("Shutting down the directory server.\n");
		Logger::Instance().Log(KLogInfo, "Shutting down the directory server.");
	}
	catch(const sydus_error & error)
	{
		printf("Unhandled exception (known): %s\n", error.what());
	}
	catch(const std::exception & error)
	{
		printf("Unhandled exception: %s\n", error.what());
	}

	return 0;
}
